<?xml version="1.0" encoding="utf-8"?>
<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.sitemaps.org/schemas/sitemap/0.9 http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd">
    {% for locale in app.project.configuration.locales %}
        {% for entity in app.project.ancestry | reject('has_generated_entity_id') | select('user_facing_entity') %}
            <url>
                <loc>{{ entity | url(absolute=true, locale=locale) }}</loc>
            </url>
        {% endfor %}
    {% endfor %}
</urlset>
