<!doctype html>
<html>
<head>
    <title>{{ app.project.configuration.title }}</title>
    <meta http-equiv="refresh" content="0; url={{ '/index.html' | url(locale=app.project.configuration.locales.default.locale) }}">
</head>
<body>
</body>
</html>
